#include <linux/log2.h>
#include <linux/slab.h>
#include "ldfs_filter.h"
#include "ldfs_store.h"

struct ddfs_proc_print_entry * print_proc_entry = NULL;

/*  Explanation
 *  We use a very compact structure in struct s_list_head
 *  It is a 64 integer
 *  However when we use slab allocator, it is aligned and the last 9 bits are unused
 *  Therefore it is used to store the record count
 *  by simplicity the last byte is used
 *
 */

static inline struct s_list_head * s_next(struct s_list_head * head) {
	return (struct s_list_head *)(head->data - (head->data & 0xFF));
}

static inline void s_next_assign(struct s_list_head * head, struct s_list_head * next) {
	head->data = (unsigned long) next - ((unsigned long) next & 0xFF) + (head->data & 0xFF);
}

static inline int s_count(struct s_list_head * head) {
	return head->data & 0xFF;
}

static inline void s_count_assign(struct s_list_head * head, int count) {
	head->data = head->data - (head->data & 0xFF) + (count & 0xFF);
}

static inline void s_list_head_init(struct s_list_head * head) {
	head->data = 0;
	s_next_assign(head, head);
}

static inline struct s_list_head * s_list_prev(struct s_list_head * head) {
	struct s_list_head * prev = head;
	for(; s_next(prev) != head; prev = s_next(prev));
	return prev;
}

static inline void s_list_add(struct s_list_head * new, struct s_list_head * head) {
	s_next_assign(new, s_next(head));
	s_next_assign(head, new);
}

static inline void s_list_del(struct s_list_head * del) {
	struct s_list_head * prev = s_list_prev(del);
	s_next_assign(prev, s_next(del));
	s_next_assign(del, NULL);
}

static inline int s_list_is_last(struct s_list_head * list, struct s_list_head * head) {
	return s_next(list) == head;
}

static inline int s_list_alone(struct s_list_head * head) {
	return s_next(head) == head;
}

#define s_list_entry(ptr, type, member) container_of(ptr, type, member)
#define s_list_for_each(pos, head, c) for(pos = head, c = 0; prefetch(s_next(head)), pos != head || c == 0; pos = s_next(pos), c = 1)
#define s_list_for_each_safe(pos, n, head) for(pos = head, n = s_next(pos); pos != (head); pos = n, n = s_next(pos))


int __ldfs_bitmap_get(struct page * ddfs_bitmap, __u32 blk_no) {
	unsigned char * ptr = (unsigned char *)page_address(ddfs_bitmap);
	unsigned char c = ptr[blk_no >> 3];
	return ((c >> (blk_no % 8)) & 1);
}

void __ldfs_bitmap_set(struct page * ddfs_bitmap, __u32 blk_no) {
	unsigned char * ptr = (unsigned char *)page_address(ddfs_bitmap);
	ptr[blk_no >> 3] |= 1 << (blk_no % 8);
}

__u32 __ldfs_index_key(struct super_block * sb, __u8 hash[]) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	__u32 key = *(__u32 *)(hash + 4);
	return key >> (32 - sbi->s_index_bit);
}

struct ldfs_bucket * ldfs_bucket_alloc(struct super_block * sb, struct s_list_head * head) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	struct ldfs_bucket * bucket = kmem_cache_alloc(sbi->s_bucket_cache, GFP_KERNEL);
	s_list_head_init(&bucket->ptr);
	if (head) {
		s_list_add(&bucket->ptr, head);
	}
	return bucket;
}

int ldfs_bucket_free(struct super_block * sb, struct s_list_head ** head, struct ldfs_bucket * bucket) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	struct s_list_head * item = &bucket->ptr;
	
	if(s_list_alone(item)) {
		*head = NULL;
	} else {
		*head = s_next(item);
	}
	s_list_del(item);
	kmem_cache_free(sbi->s_bucket_cache, bucket);
	return 0;
}

static inline __u32 ldfs_bucket_key(unsigned char * hash) {
	return (__u32)hash[0] << 24 | hash[1] << 16 | hash[2] << 8;
}

static inline void ldfs_bucket_key_assign(unsigned char * hash, __u32 key) {
	hash[0] = (key >> 24) & 0xFF;
	hash[1] = (key >> 16) & 0xFF;
	hash[2] = (key >> 8) & 0xFF;
}

int ldfs_filter_bucket_search(struct super_block * sb, struct ldfs_bucket * bucket, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no) {
	int count = s_count(&bucket->ptr);
	int left = 0, right = count, middle = 0;
	int p;
	__u32 hash_key = ldfs_bucket_key(hash);

	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		if(hash_key > ldfs_bucket_key(bucket->entry[middle].b_key)) {
			left = middle + 1;
		} else if (hash_key < ldfs_bucket_key(bucket->entry[middle].b_key)) {
			right = middle - 1;
		} else {
			while (middle > 0 && hash_key == ldfs_bucket_key(bucket->entry[middle - 1].b_key)) {
				middle--;
			}
			for (p = middle; p < count && hash_key == ldfs_bucket_key(bucket->entry[p].b_key); p++) {
				__u32 storenr = ldfs_bucket_key(bucket->entry[p].b_storenr);
				if (!ldfs_store_search(sb, hash, storenr, blk_no)) {
					return 0;
				}
			}
			break;
		}
	}
	return -1;
}

int ldfs_filter_bucket_insert(struct ldfs_bucket * bucket, __u8 hash[DDFS_HASH_LEN], __u32 blk_no) {
	int count = s_count(&bucket->ptr);
	int left = 0, right = count - 1, middle = 0;
	__u32 hash_key = ldfs_bucket_key(hash);

	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		if (hash_key > ldfs_bucket_key(bucket->entry[middle].b_key)) {
			left = middle + 1;
		} else if (hash_key < ldfs_bucket_key(bucket->entry[middle].b_key)) {
			right = middle - 1;
		} else {
			break;
		}
	}
	if (left == middle + 1) {
		middle++;
	}
	memmove(&(bucket->entry[middle + 1]), &(bucket->entry[middle]), sizeof(struct ldfs_bucket_entry) * (count - middle));
	ldfs_bucket_key_assign(bucket->entry[middle].b_key, hash_key);
	ldfs_bucket_key_assign(bucket->entry[middle].b_storenr, blk_no);
	s_count_assign(&bucket->ptr, count + 1);
	return 0;
}

int ldfs_filter_bucket_delete(struct ldfs_bucket * bucket, __u8 hash[DDFS_HASH_LEN], __u32 blk_no) {
	int count = s_count(&bucket->ptr);
	int left = 0, right = count - 1, middle = 0;
	int ret = -1;
	int entry;
	int hit = 0;
	__u32 hash_key = ldfs_bucket_key(hash);

	if (count == 0) {
		return -1;
	}
	// Binary Search
	while (left <= right) {
		middle = (left + right) / 2;
		if (hash_key > ldfs_bucket_key(bucket->entry[middle].b_key)) {
			left = middle + 1;
		} else if (hash_key < ldfs_bucket_key(bucket->entry[middle].b_key)) {
			right = middle - 1;
		} else {
			while (middle > 0 && hash_key == ldfs_bucket_key(bucket->entry[middle - 1].b_key)) {
				middle--;
			}
			hit = 1;
			break;
		}
	}
	if (!hit) {
		return ret;
	}
	for (entry = middle; entry < count && hash_key == ldfs_bucket_key(bucket->entry[entry].b_key); entry++) {
		__u32 storenr = blk_no - (blk_no & 0xFF);
		if (storenr == ldfs_bucket_key(bucket->entry[entry].b_storenr)) {
			memmove(&(bucket->entry[entry]), &(bucket->entry[entry + 1]), sizeof(struct ldfs_bucket_entry) * (count - (entry + 1)));
			s_count_assign(&bucket->ptr, count - 1);
			ret = 0;
			break;
		}
	}
	return ret;
}

extern int ldfs_load_all(struct super_block * sb);
int ldfs_filter_init(struct super_block * sb) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	unsigned page_count, i;
	struct page * ldfs_bitmap = alloc_page(GFP_KERNEL);

	// Init all mutex
	sbi->s_fingerprint_mutex = kmalloc(LDFS_LOCK_MAX * sizeof(struct mutex), GFP_KERNEL);
	for (i = 0; i < LDFS_LOCK_MAX; i++) {
		mutex_init(&sbi->s_fingerprint_mutex[i]);
	}

	if (!sbi->s_index_bit) {
		sbi->s_index_bit = ilog2(sbi->s_es->s_blocks_count) - 9;
	}

	printk(KERN_INFO "LiveDFS Index_Key Bits = %u\n", sbi->s_index_bit);
	page_count = (1 << sbi->s_index_bit);
	sbi->s_index_head = kmalloc(page_count * sizeof(struct s_list_head *), GFP_KERNEL);
	memset(sbi->s_index_head, 0, page_count * sizeof(struct s_list_head *));
    //sbi->s_bucket_cache = kmem_cache_create("ldfs_bucket_cache", sizeof(struct ldfs_bucket), 0, 0, NULL);
	sbi->s_bucket_cache = kmem_cache_create("ldfs_bucket_cache", 512, 0, 0, NULL);

/*
	for (i = 0; i < blk_cnt; i += 32768) {
		memset(page_address(ldfs_bitmap), 0, 4096);
		for (j = 0; j < 32768; j++) {
			ddfs_count_read(sb, i + j, &cnt);
			if (cnt) {
				__ldfs_bitmap_set(ldfs_bitmap, j);
			}
		}
		for (j = 0; j < 32768; j++) {
			if (__ldfs_bitmap_get(ldfs_bitmap, j)) {
				ddfs_hash_read(sb, i + j, hash);
				ldfs_filter_insert_entry(sb, hash, i + j);
			}
		}
	}
	*/
	ldfs_load_all(sb);
	__free_page(ldfs_bitmap);
	return 0;
}

int ldfs_filter_exit(struct super_block * sb) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	struct s_list_head * head;
	struct ldfs_bucket * bucket;
	int i;
	for (i = 0; i < (1 << sbi->s_index_bit); i++) {
		head = sbi->s_index_head[i];
		if (head) {
			while (!s_list_alone(head)) {
				bucket = s_list_entry(s_next(head), struct ldfs_bucket, ptr);
				ldfs_bucket_free(sb, &sbi->s_index_head[i], bucket);
			}
			bucket = s_list_entry(head, struct ldfs_bucket, ptr);
			ldfs_bucket_free(sb, &sbi->s_index_head[i], bucket);
		}
	}
	kmem_cache_destroy(sbi->s_bucket_cache);
	kfree(sbi->s_index_head);
	kfree(sbi->s_fingerprint_mutex);
	return 0;
}

int ldfs_filter_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	__u32 index = __ldfs_index_key(sb, hash);
	struct s_list_head * head = sbi->s_index_head[index];
	struct s_list_head * item;
	struct ldfs_bucket * bucket;
	int ret = -1, c;

	if (!head) {
		return -1;
	}
	s_list_for_each(item, head, c) {
		bucket = s_list_entry(item, struct ldfs_bucket, ptr);
		ret = ldfs_filter_bucket_search(sb, bucket, hash, blk_no);
		if (!ret)
			break; // Entry Found
	}
	return ret;
}

int ldfs_filter_insert_entry(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	__u32 index = __ldfs_index_key(sb, hash);
	struct s_list_head * head = sbi->s_index_head[index];
	struct s_list_head * item;
	struct ldfs_bucket * bucket;
	int ret = 0, c;

	if (head) {
		s_list_for_each(item, head, c) {
			bucket = s_list_entry(item, struct ldfs_bucket, ptr);
			if (s_count(item) < LDFS_BUCKET_MAX) {
				ldfs_filter_bucket_insert(bucket, hash, blk_no);
				return ret;
			}
		}
	}
	bucket = ldfs_bucket_alloc(sb, head);
	if(!head) {
		sbi->s_index_head[index] = &bucket->ptr;
	}
	ldfs_filter_bucket_insert(bucket, hash, blk_no);
	return ret;
}

int ldfs_filter_remove_entry(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no) {
	struct ext3_sb_info * sbi = EXT3_SB(sb);
	__u32 index = __ldfs_index_key(sb, hash);
	struct s_list_head * head = sbi->s_index_head[index];
	struct s_list_head * item;
	struct ldfs_bucket * bucket;
	int ret = -1, c;

	BUG_ON(!head);
	s_list_for_each(item, head, c) {
		BUG_ON(!item);
		bucket = s_list_entry(item, struct ldfs_bucket, ptr);
		if (!s_count(item)) {
			printk(KERN_INFO "data = %ld, head = %p, c = %d, item = %p, bucket = %p, count = %d\n", item->data, head, c, item, bucket, s_count(item));
			BUG();
		}
		ret = ldfs_filter_bucket_delete(bucket, hash, blk_no);
		if(!ret) {
			if(!s_count(item)) {
				ldfs_bucket_free(sb, &sbi->s_index_head[index], bucket);
			}
			return ret;
		}
	}
	return ret;
}

int ldfs_fingerprint_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 * blk_no) {
	return ldfs_filter_search(sb, hash, blk_no);
}

int ldfs_fingerprint_insert(handle_t * handle, struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no, __u32 new) {
	unsigned count;
	if (new) {
		ddfs_hash_write(handle, sb, blk_no, hash);
		ldfs_filter_insert_entry(sb, hash, blk_no);
	}
	ddfs_count_write_delta(handle, sb, blk_no, 1, &count);
	return 0;
}

int ldfs_fingerprint_remove(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no) {
	ldfs_filter_remove_entry(sb, hash, blk_no);
	return 0;
}


