/*
 * ddfs_store.c
 *
 *  Created on: 29 Oct 2010
 *      Author: ngchunho
 */

#include "ldfs_store.h"
#include "md5.h"
#include "ddfs.h"
#include "ext3_jbd.h"

int ddfs_hash_compute(char * dst, char * src, int len) {
	md5_state_t state;
	md5_init(&state);
	md5_append(&state, src , len);
	md5_finish(&state, dst);
	return 0;
}


enum DDFS_INDEX_TYPE{
	DDFS_INDEX_TYPE_COUNT,
	DDFS_INDEX_TYPE_HASH
};

ext3_fsblk_t ldfs_storenr_to_blocknr_region(__u32 storenr) {
	return storenr;
}

unsigned ldfs_blocknr_to_index(struct super_block * sb, ext3_fsblk_t blk_no, ext3_fsblk_t * phy_blk_no,
		unsigned * inBlockIndex, enum DDFS_INDEX_TYPE index_type) {
	unsigned long block_group;
	struct ext3_sb_info *sbi = EXT3_SB(sb);
	struct ext3_super_block * es = sbi->s_es;
	struct ext3_group_desc * desc;
	unsigned long start;
	unsigned index;
	unsigned entry_per_block = sb->s_blocksize >> 2;
	ext3_fsblk_t first_block;
	ext3_fsblk_t last_block;
	int i;
	struct buffer_head * bh;

	block_group = (blk_no - le32_to_cpu(es->s_first_data_block)) / EXT3_BLOCKS_PER_GROUP(sb);

	first_block = ext3_group_first_block_no(sb, block_group);

	BUG_ON(first_block % 32768 != 0);

	if (block_group == sbi->s_groups_count - 1)
		last_block = le32_to_cpu(sbi->s_es->s_blocks_count) - 1;
	else
		last_block = first_block + (EXT3_BLOCKS_PER_GROUP(sb) - 1);

	if (index_type == DDFS_INDEX_TYPE_HASH)
		entry_per_block = sb->s_blocksize / DDFS_HASH_LEN;

	desc = ext3_get_group_desc(sb, block_group, NULL);

	if (index_type == DDFS_INDEX_TYPE_COUNT)
		start = desc->bg_count_index_start + first_block;
	else if (index_type == DDFS_INDEX_TYPE_HASH)
		start = desc->bg_hash_index_start + first_block;
	else
		BUG_ON(1);

	index = (blk_no - first_block) / entry_per_block;
	//*inBlockIndex = (blk_no - first_block) &  (entry_per_block - 1);
	*inBlockIndex = (blk_no - first_block) % entry_per_block;

	*phy_blk_no = start + index;

	
	/* Just Try*/
	if ((bh = __find_get_block(sb->s_bdev, *phy_blk_no, PAGE_SIZE)) == NULL) {
		for (i = 0; i < ((index_type == DDFS_INDEX_TYPE_COUNT) ? 32 : 128); i++) {
			bh = sb_getblk(sb, start + i);
			if (!bh_uptodate_or_lock(bh))
				if(bh_submit_read(bh) < 0) BUG();
			
			brelse(bh);
		}
	} else {
		brelse(bh);
	}
	return 0;
}

int ddfs_count_read(struct super_block * sb, __u32 blk_no, __u32 * count) {

	ext3_fsblk_t phy_blk_no;
	unsigned inBlockIndex;
	struct buffer_head * bh;
	unsigned * counts;

	ldfs_blocknr_to_index(sb, blk_no, &phy_blk_no, &inBlockIndex, DDFS_INDEX_TYPE_COUNT);

	BUG_ON(inBlockIndex >= 1024);

	bh = sb_getblk(sb, phy_blk_no);
	if (!bh_uptodate_or_lock(bh))
		if(bh_submit_read(bh) < 0) BUG();

	counts = (unsigned*) bh->b_data;

	*count = counts[inBlockIndex];

	brelse(bh);
	return 0;
}

int ddfs_count_write_delta(handle_t * handle, struct super_block * sb, __u32 blk_no, __s32 delta, __u32 * rest_link_count) {
	ext3_fsblk_t phy_blk_no;
	unsigned inBlockIndex;
	struct buffer_head * bh;
	unsigned * counts;
	int ret = 0;

	ldfs_blocknr_to_index(sb, blk_no, &phy_blk_no, &inBlockIndex, DDFS_INDEX_TYPE_COUNT);

	bh = sb_getblk(sb, phy_blk_no);
	if (!bh_uptodate_or_lock(bh)) {
		if (bh_submit_read(bh) < 0) {
			BUG();
		}
	}

	counts = (unsigned*) bh->b_data;
	if (counts[inBlockIndex] || delta >= 0) { //Some blocks we don't care... indirect blocks..
		ext3_journal_get_write_access(handle, bh);
		if (delta >= 0 || counts[inBlockIndex] >= abs(delta)) {
			counts[inBlockIndex] += delta;
			*rest_link_count = counts[inBlockIndex];
		} else {
			BUG_ON(1);
		}
		ext3_journal_dirty_metadata(handle, bh);
	} else {
		*rest_link_count = 0;
		ret = -1;
	}
	brelse(bh);
	return ret;
}

int ddfs_count_write(handle_t * handle, struct super_block * sb, __u32 blk_no, __u32 count) {
	ext3_fsblk_t phy_blk_no;
	unsigned inBlockIndex;
	struct buffer_head * bh;
	unsigned * counts;


	ldfs_blocknr_to_index(sb,blk_no,&phy_blk_no,&inBlockIndex,DDFS_INDEX_TYPE_COUNT);
	bh = sb_getblk(sb, phy_blk_no);
	if (!bh_uptodate_or_lock(bh)) {
		if (bh_submit_read(bh) < 0) {
			BUG();
		}
	}
	ext3_journal_get_write_access(handle, bh);
	counts = (unsigned*) bh->b_data;
	counts[inBlockIndex] = count;
	ext3_journal_dirty_metadata(handle, bh);
	brelse(bh);
	return 0;
}


int ddfs_hash_read(struct super_block * sb, __u32 blk_no, __u8 hash[DDFS_HASH_LEN]) {
	ext3_fsblk_t phy_blk_no;
	unsigned inBlockIndex;
	struct buffer_head * bh;

	ldfs_blocknr_to_index(sb, blk_no, &phy_blk_no, &inBlockIndex, DDFS_INDEX_TYPE_HASH);
	bh = sb_getblk(sb, phy_blk_no);
	if (!bh_uptodate_or_lock(bh)) {
		if(bh_submit_read(bh) < 0) {
			BUG();
		}
	}
	memcpy(hash, bh->b_data + DDFS_HASH_LEN * inBlockIndex, DDFS_HASH_LEN);
	brelse(bh);

	return 0;
}

int ddfs_hash_write(handle_t * handle, struct super_block * sb, __u32 blk_no, __u8 hash[DDFS_HASH_LEN]) {
	ext3_fsblk_t phy_blk_no;
	unsigned inBlockIndex;
	struct buffer_head * bh;

	ldfs_blocknr_to_index(sb,blk_no,&phy_blk_no,&inBlockIndex,DDFS_INDEX_TYPE_HASH);
	bh = sb_getblk(sb, phy_blk_no);
	if (!bh_uptodate_or_lock(bh)) {
		if(bh_submit_read(bh) < 0) {
			BUG();
		}
	}

	ext3_journal_get_write_access(handle, bh);
	memcpy(bh->b_data + DDFS_HASH_LEN * inBlockIndex, hash, DDFS_HASH_LEN);
	ext3_journal_dirty_metadata(handle, bh);
	brelse(bh);
	return 0;
}

int ldfs_store_search(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 storenr, __u32 * blocknr) {
	ext3_fsblk_t start_blocknr = ldfs_storenr_to_blocknr_region(storenr);
	ext3_fsblk_t phy_blocknr;
	unsigned inBlockIndex;
	struct buffer_head * bh;
	int i, blknr, ct;
	ldfs_blocknr_to_index(sb, start_blocknr, &phy_blocknr, &inBlockIndex, DDFS_INDEX_TYPE_HASH);
	bh = sb_getblk(sb, phy_blocknr);
	if (!bh_uptodate_or_lock(bh)) {
		if (bh_submit_read(bh) < 0) {
			BUG();
		}
	}
	// Linear Search - 256 is 4096 / 16 (MD5 size)
	for(i = 0; i < 256; i++) {
		if (!memcmp(hash, bh->b_data + DDFS_HASH_LEN * i, DDFS_HASH_LEN)) {
			blknr = start_blocknr + i;
			ddfs_count_read(sb, blknr, &ct);
			if (ct) {
				*blocknr = blknr;
				brelse(bh);
				return 0;
			}
		}
	}
	brelse(bh);
	return -1;

}
#define LDFS_STORE_SIZE (160)
static inline ext3_fsblk_t ldfs_store_start(struct super_block * sb,unsigned block_group){
	struct ext3_group_desc * desc;
	ext3_fsblk_t first_block;
	ext3_fsblk_t start;
	desc = ext3_get_group_desc(sb, block_group, NULL);
	first_block = ext3_group_first_block_no(sb, block_group);
	start = desc->bg_hash_index_start + first_block;
	return start;
}
int _ldfs_store_prefetch_start(struct super_block * sb, ext3_fsblk_t store_start) {
	int i;
	struct buffer_head * bh;
	for (i = 0; i < LDFS_STORE_SIZE; i++) {
		bh = sb_getblk(sb, store_start + i);
		if(buffer_uptodate(bh)) {
			continue;
		}
		lock_buffer(bh);
		get_bh(bh);
		bh->b_end_io = end_buffer_read_sync;
		submit_bh(READ, bh);
		brelse(bh);
	
	}
	return 0;
}
extern int ldfs_filter_insert_entry(struct super_block * sb, __u8 hash[DDFS_HASH_LEN], __u32 blk_no);
int ldfs_load_all(struct super_block * sb){
	int block_group = 0;
	unsigned long block_no = 0;
	struct ext3_sb_info *sbi = EXT3_SB(sb);
	struct ext3_group_desc * desc;
	int ii,jj;
	struct buffer_head * bh;
	struct buffer_head * bh_hash = NULL;
	ext3_fsblk_t first_block;
	ext3_fsblk_t count_start;
	ext3_fsblk_t hash_blk;
	ext3_fsblk_t hash_cur = ~0;

	_ldfs_store_prefetch_start(sb, ldfs_store_start(sb,0));
	for(block_group = 0; block_group < sbi->s_groups_count; block_group += 1){
		desc = ext3_get_group_desc(sb, block_group, NULL);
		first_block = ext3_group_first_block_no(sb, block_group);
		count_start = desc->bg_count_index_start + first_block;
		hash_blk = desc->bg_hash_index_start + first_block;


		if(block_group + 1 < sbi->s_groups_count){
			_ldfs_store_prefetch_start(sb, ldfs_store_start(sb,block_group + 1));
		}
		for(ii = 0; ii < 128;ii++){
			unsigned * counts;
			bh = sb_getblk(sb, count_start+ii);
			if (!buffer_uptodate(bh))
			{
				wait_on_buffer(bh);
			}

			counts = (unsigned*) bh->b_data;

			hash_blk--; // ^_^ For saving a "jj > 0" condition in the "if"
			for(jj = 0;jj < 1024; jj++,block_no++){
				if(jj % (4096 / DDFS_HASH_LEN) == 0){
					hash_blk++;
				}

				if(counts[jj] != 0){
					unsigned index = jj % 256;
					if(hash_cur != hash_blk){
						if(bh_hash){
							brelse(bh_hash);
						}
						bh_hash = sb_getblk(sb,hash_blk);
						if (!buffer_uptodate(bh_hash))
						{
							wait_on_buffer(bh_hash);
						}
						hash_cur = hash_blk;
					}
					ldfs_filter_insert_entry(sb, bh_hash->b_data + index*DDFS_HASH_LEN , block_no);
				}
			}
			brelse(bh);
		}
	}
	if(bh_hash){
		brelse(bh_hash);
	}
	return 0;

}
